@@ -36,7 +36,7 @@ size_t get_wkt_split(struct osmNode *, int count, int polygon, double split_at);
 
 char* get_wkt(size_t index);
 double get_area(size_t index);
-size_t build_geometry(int osm_id, struct osmNode **xnodes, int *xcount, int make_polygon, int enable_multi, double split_at);
+size_t build_geometry(osmid_t osm_id, struct osmNode **xnodes, int *xcount, int make_polygon, int enable_multi, double split_at);
 void clear_wkts();
 
 #ifdef __cplusplus
