@@ -21,8 +21,8 @@
 
 #define EARTH_CIRCUMFERENCE		40075016.68
 #define HALF_EARTH_CIRCUMFERENCE	(EARTH_CIRCUMFERENCE / 2)
-#define TILE_EXPIRY_LEEWAY		0.5		// How many tiles worth of space to leave either side of a changed feature
-#define EXPIRE_TILES_MAX_BBOX		30000		// Maximum width or height of a bounding box (metres)
+#define TILE_EXPIRY_LEEWAY		0.1		// How many tiles worth of space to leave either side of a changed feature
+#define EXPIRE_TILES_MAX_BBOX		20000		// Maximum width or height of a bounding box (metres)
 
 struct tile {
 	int		complete[2][2];	// Flags
@@ -263,9 +263,9 @@ static void expire_tiles_from_line(double lon_a, double lat_a, double lon_b, dou
 	y_step = y_len / hyp_len;
 //	fprintf(stderr, "Expire from line (%f,%f),(%f,%f) [%f,%f],[%f,%f] %fx%f hyp_len = %f\n", lon_a, lat_a, lon_b, lat_b, tile_x_a, tile_y_a, tile_x_b, tile_y_b, x_len, y_len, hyp_len);
 	
-	for (step = 0; step <= hyp_len; step ++) {
+	for (step = 0; step <= hyp_len; step+= 0.4) {
 		// Interpolate points 1 tile width apart
-		next_step = step + 1;
+		next_step = step + 0.4;
 		if (next_step > hyp_len) next_step = hyp_len;
 		x1 = tile_x_a + ((double)step * x_step);
 		y1 = tile_y_a + ((double)step * y_step);
@@ -371,7 +371,7 @@ void expire_tiles_from_nodes_line(struct osmNode * nodes, int count) {
 /*
  * Calculate a bounding box from a list of nodes and expire all tiles within it
  */
-void expire_tiles_from_nodes_poly(struct osmNode * nodes, int count, int osm_id) {
+void expire_tiles_from_nodes_poly(struct osmNode * nodes, int count, osmid_t osm_id) {
 	int	i;
 	int	got_coords = 0;
 	double	min_lon = 0.0;
@@ -391,13 +391,13 @@ void expire_tiles_from_nodes_poly(struct osmNode * nodes, int count, int osm_id)
 	if (got_coords) {
 		if (expire_tiles_from_bbox(min_lon, min_lat, max_lon, max_lat)) {
 			// Bounding box too big - just expire tiles on the line
-			fprintf(stderr, "\rLarge polygon (%.0f x %.0f metres, OSM ID %i) - only expiring perimeter\n", max_lon - min_lon, max_lat - min_lat, osm_id);
+			fprintf(stderr, "\rLarge polygon (%.0f x %.0f metres, OSM ID %" PRIdOSMID ") - only expiring perimeter\n", max_lon - min_lon, max_lat - min_lat, osm_id);
 			expire_tiles_from_nodes_line(nodes, count);
 		}
 	}
 }
 
-static void expire_tiles_from_xnodes_poly(struct osmNode ** xnodes, int * xcount, int osm_id) {
+static void expire_tiles_from_xnodes_poly(struct osmNode ** xnodes, int * xcount, osmid_t osm_id) {
 	int	i;
 
         for (i = 0; xnodes[i]; i++) expire_tiles_from_nodes_poly(xnodes[i], xcount[i], osm_id);
@@ -409,7 +409,7 @@ static void expire_tiles_from_xnodes_line(struct osmNode ** xnodes, int * xcount
         for (i = 0; xnodes[i]; i++) expire_tiles_from_nodes_line(xnodes[i], xcount[i]);
 }
 
-void expire_tiles_from_wkt(const char * wkt, int osm_id) {
+void expire_tiles_from_wkt(const char * wkt, osmid_t osm_id) {
 	struct osmNode **	xnodes;
 	int *			xcount;
 	int			polygon;
@@ -425,22 +425,39 @@ void expire_tiles_from_wkt(const char * wkt, int osm_id) {
 	}
 }
 
-void expire_tiles_from_db(PGconn * sql_conn, int osm_id) {
-	PGresult *	res;
-	char		tmp[16];
-	char const *	paramValues[1];
-	int		tuple;
-	char *		wkt;
+/*
+ * Expire tiles based on an osm element.
+ * What type of element (node, line, polygon) osm_id refers to depends on
+ * sql_conn. Each type of table has its own sql_conn and the prepared statement
+ * get_wkt refers to the appropriate table.
+ *
+ * The function returns -1 if expiry is not enabled. Otherwise it returns the number
+ * of elements that refer to the osm_id.
 
-	if (Options->expire_tiles_zoom < 0) return;
-	snprintf(tmp, sizeof(tmp), "%d", osm_id);
-	paramValues[0] = tmp;
-	res = pgsql_execPrepared(sql_conn, "get_way", 1, paramValues, PGRES_TUPLES_OK);
-	for (tuple = 0; tuple < PQntuples(res); tuple++) {
-		wkt = PQgetvalue(res, tuple, 0);
-		expire_tiles_from_wkt(wkt, osm_id);
-	}
-	PQclear(res);
+ */
+int expire_tiles_from_db(PGconn * sql_conn, osmid_t osm_id) {
+    PGresult *	res;
+    char *		wkt;
+    int i, noElements = 0;
+    char const *paramValues[1];
+    char tmp[16];
+
+    if (Options->expire_tiles_zoom < 0) return -1;
+    snprintf(tmp, sizeof(tmp), "%" PRIdOSMID, osm_id);
+    paramValues[0] = tmp;
+    
+    /* The prepared statement get_wkt will behave differently depending on the sql_conn
+     * each table has its own sql_connection with the get_way refering to the approriate table
+     */
+    res = pgsql_execPrepared(sql_conn, "get_wkt", 1, (const char * const *)paramValues, PGRES_TUPLES_OK);
+    noElements = PQntuples(res);
+
+    for (i = 0; i < noElements; i++) {
+        wkt = PQgetvalue(res, i, 0);
+        expire_tiles_from_wkt(wkt, osm_id);
+    }
+    PQclear(res);
+    return noElements;
 }
 
 
