@@ -25,6 +25,7 @@ void initList(struct keyval *head)
     head->prev = head;
     head->key = NULL;
     head->value = NULL;
+    head->has_column = 0;
 }
 
 void freeItem(struct keyval *p)
@@ -84,6 +85,33 @@ char *getItem(struct keyval *head, const char *name)
     return NULL;
 }	
 
+/* unlike getItem this function gives a pointer to the whole
+   list item which can be used to remove the tag from the linked list
+   with the removeTag function
+*/
+struct keyval *getTag(struct keyval *head, const char *name)
+{
+    struct keyval *p;
+
+    if (!head) 
+        return NULL;
+
+    p = head->next;
+    while(p != head) {
+        if (!strcmp(p->key, name))
+            return p;
+        p = p->next;
+    }
+    return NULL;
+}
+
+void removeTag(struct keyval *tag)
+{
+  tag->prev->next=tag->next;
+  tag->next->prev=tag->prev;
+  freeItem(tag);
+}
+
 struct keyval *firstItem(struct keyval *head)
 {
     if (head == NULL || head == head->next)
@@ -183,6 +211,7 @@ struct keyval *popItem(struct keyval *head)
 
 void pushItem(struct keyval *head, struct keyval *item)
 {
+
     assert(head);
     assert(item);
  
@@ -223,6 +252,8 @@ int addItem(struct keyval *head, const char *name, const char *value, int noDupe
     item->key   = strdup(name);
     item->value = strdup(value);
 #endif
+    item->has_column=0;
+
 
 #if 1
     // Add to head
