@@ -17,6 +17,7 @@
 #include "osmtypes.h"
 #include "middle.h"
 #include "middle-ram.h"
+#include "node-ram-cache.h"
 
 #include "output-pgsql.h"
 
@@ -25,22 +26,10 @@
 #define FIXED_POINT
 
 static int scale = 100;
-#define DOUBLE_TO_FIX(x) ((x) * scale)
-#define FIX_TO_DOUBLE(x) (((double)x) / scale)
-
-struct ramNode {
-#ifdef FIXED_POINT
-    int lon;
-    int lat;
-#else
-    double lon;
-    double lat;
-#endif
-};
 
 struct ramWay {
     struct keyval *tags;
-    int *ndids;
+    osmid_t *ndids;
     int pending;
 };
 
@@ -64,11 +53,10 @@ struct ramRel {
  *
  */
 
-#define BLOCK_SHIFT 14
+#define BLOCK_SHIFT 10
 #define PER_BLOCK  (1 << BLOCK_SHIFT)
 #define NUM_BLOCKS (1 << (32 - BLOCK_SHIFT))
 
-static struct ramNode    *nodes[NUM_BLOCKS];
 static struct ramWay     *ways[NUM_BLOCKS];
 static struct ramRel     *rels[NUM_BLOCKS];
 
@@ -78,13 +66,13 @@ static int way_blocks;
 static int way_out_count;
 static int rel_out_count;
 
-static inline int id2block(int id)
+static inline osmid_t id2block(osmid_t id)
 {
     // + NUM_BLOCKS/2 allows for negative IDs
     return (id >> BLOCK_SHIFT) + NUM_BLOCKS/2;
 }
 
-static inline int id2offset(int id)
+static inline osmid_t id2offset(osmid_t id)
 {
     return id & (PER_BLOCK-1);
 }
@@ -95,54 +83,8 @@ static inline int block2id(int block, int offset)
 }
 
 #define UNUSED  __attribute__ ((unused))
-static int ram_nodes_set(int id, double lat, double lon, struct keyval *tags UNUSED)
-{
-    int block  = id2block(id);
-    int offset = id2offset(id);
 
-    if (!nodes[block]) {
-        nodes[block] = calloc(PER_BLOCK, sizeof(struct ramNode));
-        if (!nodes[block]) {
-            fprintf(stderr, "Error allocating nodes\n");
-            exit_nicely();
-        }
-        node_blocks++;
-        //fprintf(stderr, "\tnodes(%zuMb)\n", node_blocks * sizeof(struct ramNode) * PER_BLOCK / 1000000);
-    }
-
-#ifdef FIXED_POINT
-    nodes[block][offset].lat = DOUBLE_TO_FIX(lat);
-    nodes[block][offset].lon = DOUBLE_TO_FIX(lon);
-#else
-    nodes[block][offset].lat = lat;
-    nodes[block][offset].lon = lon;
-#endif
-    return 0;
-}
-
-
-static int ram_nodes_get(struct osmNode *out, int id)
-{
-    int block  = id2block(id);
-    int offset = id2offset(id);
-
-    if (!nodes[block])
-        return 1;
-
-    if (!nodes[block][offset].lat && !nodes[block][offset].lon)
-        return 1;
-
-#ifdef FIXED_POINT
-    out->lat = FIX_TO_DOUBLE(nodes[block][offset].lat);
-    out->lon = FIX_TO_DOUBLE(nodes[block][offset].lon);
-#else
-    out->lat = nodes[block][offset].lat;
-    out->lon = nodes[block][offset].lon;
-#endif
-    return 0;
-}
-
-static int ram_ways_set(int id, int *nds, int nd_count, struct keyval *tags, int pending)
+static int ram_ways_set(osmid_t id, osmid_t *nds, int nd_count, struct keyval *tags, int pending)
 {
     int block  = id2block(id);
     int offset = id2offset(id);
@@ -164,8 +106,8 @@ static int ram_ways_set(int id, int *nds, int nd_count, struct keyval *tags, int
     }
 
     /* Copy into length prefixed array */
-    ways[block][offset].ndids = malloc( (nd_count+1)*sizeof(int) );
-    memcpy( ways[block][offset].ndids+1, nds, nd_count*sizeof(int) );
+    ways[block][offset].ndids = malloc( (nd_count+1)*sizeof(osmid_t) );
+    memcpy( ways[block][offset].ndids+1, nds, nd_count*sizeof(osmid_t) );
     ways[block][offset].ndids[0] = nd_count;
     ways[block][offset].pending = pending;
 
@@ -186,7 +128,7 @@ static int ram_ways_set(int id, int *nds, int nd_count, struct keyval *tags, int
     return 0;
 }
 
-static int ram_relations_set(int id, struct member *members, int member_count, struct keyval *tags)
+static int ram_relations_set(osmid_t id, struct member *members, int member_count, struct keyval *tags)
 {
     struct keyval *p;
     int block  = id2block(id);
@@ -229,14 +171,14 @@ static int ram_relations_set(int id, struct member *members, int member_count, s
     return 0;
 }
 
-static int ram_nodes_get_list(struct osmNode *nodes, int *ndids, int nd_count)
+static int ram_nodes_get_list(struct osmNode *nodes, osmid_t *ndids, int nd_count)
 {
     int i, count;
 
     count = 0;
     for( i=0; i<nd_count; i++ )
     {
-        if (ram_nodes_get(&nodes[count], ndids[i]))
+        if (ram_cache_nodes_get(&nodes[count], ndids[i]))
             continue;
 
         count++;
@@ -244,7 +186,7 @@ static int ram_nodes_get_list(struct osmNode *nodes, int *ndids, int nd_count)
     return count;
 }
 
-static void ram_iterate_relations(int (*callback)(int id, struct member *members, int member_count, struct keyval *tags, int))
+static void ram_iterate_relations(int (*callback)(osmid_t id, struct member *members, int member_count, struct keyval *tags, int))
 {
     int block, offset;
 
@@ -255,7 +197,7 @@ static void ram_iterate_relations(int (*callback)(int id, struct member *members
 
         for (offset=0; offset < PER_BLOCK; offset++) {
             if (rels[block][offset].members) {
-                int id = block2id(block, offset);
+                osmid_t id = block2id(block, offset);
                 rel_out_count++;
                 if (rel_out_count % 10 == 0)
                     fprintf(stderr, "\rWriting relation (%u)", rel_out_count);
@@ -275,7 +217,7 @@ static void ram_iterate_relations(int (*callback)(int id, struct member *members
     fprintf(stderr, "\rWriting relation (%u)\n", rel_out_count);
 }
 
-static void ram_iterate_ways(int (*callback)(int id, struct keyval *tags, struct osmNode *nodes, int count, int exists))
+static void ram_iterate_ways(int (*callback)(osmid_t id, struct keyval *tags, struct osmNode *nodes, int count, int exists))
 {
     int block, offset, ndCount = 0;
     struct osmNode *nodes;
@@ -297,7 +239,7 @@ static void ram_iterate_ways(int (*callback)(int id, struct keyval *tags, struct
                     ndCount = ram_nodes_get_list(nodes, ways[block][offset].ndids+1, ways[block][offset].ndids[0]);
 
                     if (nodes) {
-                        int id = block2id(block, offset);
+                        osmid_t id = block2id(block, offset);
                         callback(id, ways[block][offset].tags, nodes, ndCount, 0);
                         free(nodes);
                     }
@@ -321,7 +263,7 @@ static void ram_iterate_ways(int (*callback)(int id, struct keyval *tags, struct
 }
 
 /* Caller must free nodes_ptr and resetList(tags_ptr) */
-static int ram_ways_get( int id, struct keyval *tags_ptr, struct osmNode **nodes_ptr, int *count_ptr)
+static int ram_ways_get(osmid_t id, struct keyval *tags_ptr, struct osmNode **nodes_ptr, int *count_ptr)
 {
     int block = id2block(id), offset = id2offset(id), ndCount = 0;
     struct osmNode *nodes;
@@ -345,7 +287,7 @@ static int ram_ways_get( int id, struct keyval *tags_ptr, struct osmNode **nodes
 }
 
 // Marks the way so that iterate ways skips it
-static int ram_ways_done( int id )
+static int ram_ways_done(osmid_t id)
 {
     int block = id2block(id), offset = id2offset(id);
 
@@ -372,6 +314,8 @@ static int ram_start(const struct output_options *options)
     // The fixed poing scaling needs adjusting accordingly to
     // be stored accurately in an int
     scale = options->scale;
+
+    init_node_ram_cache( options->alloc_chunkwise, options->cache, scale);
     
     fprintf( stderr, "Mid: Ram, scale=%d\n", scale );
 
@@ -381,12 +325,9 @@ static int ram_start(const struct output_options *options)
 static void ram_stop(void)
 {
     int i, j;
+    free_node_ram_cache();
 
     for (i=0; i<NUM_BLOCKS; i++) {
-        if (nodes[i]) {
-            free(nodes[i]);
-            nodes[i] = NULL;
-        }
         if (ways[i]) {
             for (j=0; j<PER_BLOCK; j++) {
                 if (ways[i][j].tags) {
@@ -402,13 +343,17 @@ static void ram_stop(void)
     }
 }
 
+static void ram_commit(void) {
+}
+
 struct middle_t mid_ram = {
     .start             = ram_start,
     .stop              = ram_stop,
     .end               = ram_end,
     .cleanup           = ram_stop,
     .analyze           = ram_analyze,
-    .nodes_set         = ram_nodes_set,
+    .commit            = ram_commit,
+    .nodes_set         = ram_cache_nodes_set,
 #if 0
     .nodes_get         = ram_nodes_get,
 #endif
