@@ -44,6 +44,7 @@
 #include "build_geometry.h"
 #include "middle-pgsql.h"
 #include "middle-ram.h"
+#include "node-ram-cache.h"
 #include "output-pgsql.h"
 #include "output-gazetteer.h"
 #include "output-null.h"
@@ -133,9 +134,16 @@ static void long_usage(char *arg0)
     printf("   -c|--create\t\tRemove existing data from the database. This is the \n");
     printf("              \t\tdefault if --append is not specified.\n");
     printf("   -d|--database\tThe name of the PostgreSQL database to connect\n");
-    printf("                \tto (default: gis).\n");
-    printf("   -i|--tablespace-index\tThe name of the PostgreSQL tablespace where indexes will be create\n");
-    printf("                \tto (default: pg_default).\n");
+    printf("              \t\tto (default: gis).\n");
+    printf("   -i|--tablespace-index\tThe name of the PostgreSQL tablespace where\n");
+    printf("              \t\tall indexes will be created.\n");
+    printf("              \t\tThe following options allow more fine-grained control:\n");
+    printf("      --tablespace-main-data \ttablespace for main tables\n");
+    printf("      --tablespace-main-index\ttablespace for main table indexes\n");
+    printf("      --tablespace-slim-data \ttablespace for slim mode tables\n");
+    printf("      --tablespace-slim-index\ttablespace for slim mode indexes\n");
+    printf("              \t\t(if unset, use db's default; -i is equivalent to setting\n");
+    printf("              \t\t--tablespace-main-index and --tablespace-slim-index)\n");
     printf("   -l|--latlong\t\tStore data in degrees of latitude & longitude.\n");
     printf("   -m|--merc\t\tStore data in proper spherical mercator (default)\n");
     printf("   -M|--oldmerc\t\tStore data in the legacy OSM mercator format\n");
@@ -144,18 +152,21 @@ static void long_usage(char *arg0)
     printf("                \tdumps prior to August 2007). Adds about 10%% overhead.\n");
     printf("   -p|--prefix\t\tPrefix for table names (default planet_osm)\n");
     printf("   -s|--slim\t\tStore temporary data in the database. This greatly\n");
-    printf("            \t\treduces the RAM usage but is much slower.\n");
+    printf("            \t\treduces the RAM usage but is much slower. This switch is\n");
+    printf("            \t\trequired if you want to update with --append later.\n");
 
     if (sizeof(int*) == 4) {
-        printf("            \t\tYou are running this on 32bit system, so at most\n");
-        printf("            \t\t3GB of RAM will be used. If you encounter unexpected\n");
-        printf("            \t\texceptions during import, you should try this switch.\n");
+        printf("            \t\tThis program was compiled on a 32bit system, so at most\n");
+        printf("            \t\t3GB of RAM will be used. If you encounter problems\n");
+        printf("            \t\tduring import, you should try this switch.\n");
     }
+    printf("      --drop\t\tonly with --slim: drop temporary tables after import (no updates).\n");
     
     printf("   -S|--style\t\tLocation of the style file. Defaults to " OSM2PGSQL_DATADIR "/default.style\n");
-    printf("   -C|--cache\t\tOnly for slim mode: Use upto this many MB for caching nodes\n");
-    printf("             \t\tDefault is 800\n");
-    printf("   -U|--username\tPostgresql user name.\n");
+    printf("   -C|--cache\t\tNow required for slim and non-slim modes: \n");
+    printf("             \t\tUse up to this many MB for caching nodes (default: 800)\n");
+    printf("   -U|--username\tPostgresql user name\n");
+    printf("             \t\tpassword can be given by prompt or PGPASS environment variable.\n");
     printf("   -W|--password\tForce password prompt.\n");
     printf("   -H|--host\t\tDatabase server hostname or socket location.\n");
     printf("   -P|--port\t\tDatabase server port.\n");
@@ -175,14 +186,33 @@ static void long_usage(char *arg0)
     printf("              \t\tInclude attributes for each object in the database.\n");
     printf("              \t\tThis includes the username, userid, timestamp and version.\n"); 
     printf("              \t\tNote: this option also requires additional entries in your style file.\n"); 
-    printf("   -k|--hstore\t\tGenerate an additional hstore (key/value) column to  postgresql tables\n");
-    printf("   -z|--hstore-column\tGenerate an additional hstore (key/value) column to containing all tags\n");
+    printf("   -k|--hstore\t\tAdd tags without column to an additional hstore (key/value) column to postgresql tables\n");
+    printf("   -j|--hstore-all\tAdd all tags to an additional hstore (key/value) column in postgresql tables\n");
+    printf("   -z|--hstore-column\tAdd an additional hstore (key/value) column containing all tags\n");
     printf("                     \tthat start with the specified string, eg --hstore-column \"name:\" will\n");
     printf("                     \tproduce an extra hstore column that contains all name:xx tags\n");
     printf("   -G|--multi-geometry\tGenerate multi-geometry features in postgresql tables.\n");
     printf("   -K|--keep-coastlines\tKeep coastline data rather than filtering it out.\n");
     printf("              \t\tBy default natural=coastline tagged data will be discarded based on the\n");
     printf("              \t\tassumption that post-processed Coastline Checker shapefiles will be used.\n");
+    printf("      --number-processes\t\tSpecifies the number of parallel processes used for certain operations\n");
+    printf("             \t\tDefault is 1\n");
+    printf("   -I|--disable-parallel-indexing\tDisable indexing all tables concurrently.\n");
+    printf("      --unlogged-tables\tUse unlogged tables (lost on crash but faster). Requires PostgreSQL 9.1.\n");
+    printf("      --cache-strategy\tSpecifies the method used to cache nodes in ram.\n");
+    printf("                      \t\tAvailable options are:\n");
+    printf("                      \t\tdense: caching strategy optimised for full planet import\n");
+    printf("                      \t\tchunked: caching strategy optimised for non-contigouse memory allocation\n");
+    printf("                      \t\tsparse: caching strategy optimised for small extracts\n");
+    printf("                      \t\toptimized: automatically combines dense and sparse strategies for optimal storage efficiency.\n");
+    printf("                      \t\t           optimized may use twice as much virtual memory, but no more physical memory\n");
+#ifdef __amd64__
+    printf("                      \t\t   The default is \"optimized\"\n");
+#else
+    /* use "chunked" as a default in 32 bit compilations, as it is less wasteful of virtual memory than "optimized"*/
+    printf("                      \t\t   The default is \"chunked\"\n");
+#endif
+
     printf("   -h|--help\t\tHelp information.\n");
     printf("   -v|--verbose\t\tVerbose output.\n");
     printf("\n");
@@ -273,8 +303,20 @@ void resetMembers(struct osmdata_t *osmdata)
 
 void printStatus(struct osmdata_t *osmdata)
 {
-    fprintf(stderr, "\rProcessing: Node(%dk) Way(%dk) Relation(%d)",
-            osmdata->count_node/1000, osmdata->count_way/1000, osmdata->count_rel);
+    time_t now;
+    time(&now);
+    time_t end_nodes = osmdata->start_way > 0 ? osmdata->start_way : now;
+    time_t end_way = osmdata->start_rel > 0 ? osmdata->start_rel : now;
+    time_t end_rel =  now;
+    fprintf(stderr, "\rProcessing: Node(%" PRIdOSMID "k %.1fk/s) Way(%" PRIdOSMID "k %.2fk/s) Relation(%" PRIdOSMID " %.2f/s)",
+            osmdata->count_node/1000,
+            (double)osmdata->count_node/1000.0/((int)(end_nodes - osmdata->start_node) > 0 ? (double)(end_nodes - osmdata->start_node) : 1.0),
+            osmdata->count_way/1000,
+            osmdata->count_way > 0 ? (double)osmdata->count_way/1000.0/
+            ((double)(end_way - osmdata->start_way) > 0.0 ? (double)(end_way - osmdata->start_way) : 1.0) : 0.0,
+            osmdata->count_rel,
+            osmdata->count_rel > 0 ? (double)osmdata->count_rel/
+            ((double)(end_rel - osmdata->start_rel) > 0.0 ? (double)(end_rel - osmdata->start_rel) : 1.0) : 0.0);
 }
 
 int node_wanted(struct osmdata_t *osmdata, double lat, double lon)
@@ -300,15 +342,27 @@ int main(int argc, char *argv[])
     int projection = PROJ_SPHERE_MERC;
     int expire_tiles_zoom = -1;
     int expire_tiles_zoom_min = -1;
-    int enable_hstore = 0;
+    int enable_hstore = HSTORE_NONE;
     int enable_multi = 0;
+    int parallel_indexing = 1;
+#ifdef __amd64__
+    int alloc_chunkwise = ALLOC_SPARSE | ALLOC_DENSE;
+#else
+    int alloc_chunkwise = ALLOC_DENSE_CHUNK | ALLOC_DENSE;
+#endif
+    int num_procs = 1;
+    int droptemp = 0;
+    int unlogged = 0;
     const char *expire_tiles_filename = "dirty_tiles";
     const char *db = "gis";
     const char *username=NULL;
     const char *host=NULL;
     const char *password=NULL;
     const char *port = "5432";
-    const char *tblsindex = "pg_default"; // default TABLESPACE for index
+    const char *tblsmain_index = NULL; // no default TABLESPACE for index on main tables
+    const char *tblsmain_data = NULL;  // no default TABLESPACE for main tables
+    const char *tblsslim_index = NULL; // no default TABLESPACE for index on slim mode tables
+    const char *tblsslim_data = NULL;  // no default TABLESPACE for slim mode tables
     const char *conninfo = NULL;
     const char *prefix = "planet_osm";
     const char *style = OSM2PGSQL_DATADIR "/default.style";
@@ -324,7 +378,7 @@ int main(int argc, char *argv[])
     
     int (*streamFile)(char *, int, struct osmdata_t *);
 
-    printf("osm2pgsql SVN version %s\n\n", VERSION);
+    printf("osm2pgsql SVN version %s (%lubit id space)\n\n", VERSION, 8 * sizeof(osmid_t));
 
     while (1) {
         int c, option_index = 0;
@@ -347,6 +401,10 @@ int main(int argc, char *argv[])
             {"host",     1, 0, 'H'},
             {"port",     1, 0, 'P'},
             {"tablespace-index", 1, 0, 'i'},
+            {"tablespace-slim-data", 1, 0, 200},
+            {"tablespace-slim-index", 1, 0, 201},
+            {"tablespace-main-data", 1, 0, 202},
+            {"tablespace-main-index", 1, 0, 203},
             {"help",     0, 0, 'h'},
             {"style",    1, 0, 'S'},
             {"expire-tiles", 1, 0, 'e'},
@@ -354,15 +412,21 @@ int main(int argc, char *argv[])
             {"output",   1, 0, 'O'},
             {"extra-attributes", 0, 0, 'x'},
             {"hstore", 0, 0, 'k'},
+            {"hstore-all", 0, 0, 'j'},
             {"hstore-column", 1, 0, 'z'},
             {"multi-geometry", 0, 0, 'G'},
             {"keep-coastlines", 0, 0, 'K'},
             {"input-reader", 1, 0, 'r'},
             {"version", 0, 0, 'V'},
+            {"disable-parallel-indexing", 0, 0, 'I'},
+            {"cache-strategy", 1, 0, 204},
+            {"number-processes", 1, 0, 205},
+            {"drop", 0, 0, 206},
+            {"unlogged", 0, 0, 207},
             {0, 0, 0, 0}
         };
 
-        c = getopt_long (argc, argv, "ab:cd:KhlmMp:suvU:WH:P:i:E:C:S:e:o:O:xkGz:r:V", long_options, &option_index);
+        c = getopt_long (argc, argv, "ab:cd:KhlmMp:suvU:WH:P:i:IE:C:S:e:o:O:xkjGz:r:V", long_options, &option_index);
         if (c == -1)
             break;
 
@@ -386,26 +450,45 @@ int main(int argc, char *argv[])
             case 'H': host=optarg; break;
             case 'P': port=optarg; break;
             case 'S': style=optarg; break;
-            case 'i': tblsindex=optarg; break;
+            case 'i': tblsmain_index=tblsslim_index=optarg; break;
+            case 200: tblsslim_data=optarg; break;    
+            case 201: tblsslim_index=optarg; break;    
+            case 202: tblsmain_data=optarg; break;    
+            case 203: tblsmain_index=optarg; break;    
             case 'e':
                 expire_tiles_zoom_min = atoi(optarg);
-		temparg = strchr(optarg, '-');
-		if (temparg) expire_tiles_zoom = atoi(temparg + 1);
-		if (expire_tiles_zoom < expire_tiles_zoom_min) expire_tiles_zoom = expire_tiles_zoom_min;
+                temparg = strchr(optarg, '-');
+                if (temparg) expire_tiles_zoom = atoi(temparg + 1);
+                if (expire_tiles_zoom < expire_tiles_zoom_min) expire_tiles_zoom = expire_tiles_zoom_min;
                 break;
             case 'o': expire_tiles_filename=optarg; break;
-	    case 'O': output_backend = optarg; break;
+            case 'O': output_backend = optarg; break;
             case 'x': osmdata.extra_attributes=1; break;
-	    case 'k': enable_hstore=1; break;
+            case 'k': enable_hstore=HSTORE_NORM; break;
+	    case 'j': enable_hstore=HSTORE_ALL; break;
             case 'z': 
                 n_hstore_columns++;
                 hstore_columns = (const char**)realloc(hstore_columns, sizeof(&n_hstore_columns) * n_hstore_columns);
                 hstore_columns[n_hstore_columns-1] = optarg;
                 break;
-	    case 'G': enable_multi=1; break;
-	    case 'r': input_reader = optarg; break;
+            case 'G': enable_multi=1; break;
+            case 'r': input_reader = optarg; break;
             case 'h': long_usage_bool=1; break;
-  	    case 'V': exit(EXIT_SUCCESS);
+            case 'I': 
+#ifdef HAVE_PTHREAD
+                parallel_indexing=0; 
+#endif
+                break;
+            case 204:
+                if (strcmp(optarg,"dense") == 0) alloc_chunkwise = ALLOC_DENSE;
+                if (strcmp(optarg,"chunk") == 0) alloc_chunkwise = ALLOC_DENSE | ALLOC_DENSE_CHUNK;
+                if (strcmp(optarg,"sparse") == 0) alloc_chunkwise = ALLOC_SPARSE;
+                if (strcmp(optarg,"optimized") == 0) alloc_chunkwise = ALLOC_DENSE | ALLOC_SPARSE;
+                break;
+            case 205: num_procs = atoi(optarg); break;
+            case 206: droptemp = 1; break;
+            case 207: unlogged = 1; break;
+            case 'V': exit(EXIT_SUCCESS);
             case '?':
             default:
                 short_usage(argv[0]);
@@ -428,21 +511,38 @@ int main(int argc, char *argv[])
         exit(EXIT_FAILURE);
     }
 
-    if( cache < 0 ) cache = 0;
+    if (droptemp && !slim) {
+        fprintf(stderr, "Error: --drop only makes sense with --slim.\n");
+        exit(EXIT_FAILURE);
+    }
+
+    if (unlogged && !create) {
+        fprintf(stderr, "Warning: --unlogged only makes sense with --create; ignored.\n");
+        unlogged = 0;
+    }
+
+    if (cache < 0) cache = 0;
+
+    if (num_procs < 1) num_procs = 1;
 
     if (pass_prompt)
         password = simple_prompt("Password:", 100, 0);
     else {
         password = getenv("PGPASS");
     }	
-        
 
     conninfo = build_conninfo(db, username, password, host, port);
     sql_conn = PQconnectdb(conninfo);
     if (PQstatus(sql_conn) != CONNECTION_OK) {
-        fprintf(stderr, "Connection to database failed: %s\n", PQerrorMessage(sql_conn));
+        fprintf(stderr, "Error: Connection to database failed: %s\n", PQerrorMessage(sql_conn));
         exit(EXIT_FAILURE);
     }
+    if (unlogged && PQserverVersion(sql_conn) < 90100) {
+        fprintf(stderr, "Error: --unlogged works only with PostgreSQL 9.1 and above, but\n");
+        fprintf(stderr, "you are using PostgreSQL %d.%d.%d.\n", PQserverVersion(sql_conn) / 10000, (PQserverVersion(sql_conn) / 100) % 100, PQserverVersion(sql_conn) % 100);
+        exit(EXIT_FAILURE);
+    }
+
     PQfinish(sql_conn);
 
     text_init();
@@ -470,7 +570,10 @@ int main(int argc, char *argv[])
     options.mid = slim ? &mid_pgsql : &mid_ram;
     options.cache = cache;
     options.style = style;
-    options.tblsindex = tblsindex;
+    options.tblsmain_index = tblsmain_index;
+    options.tblsmain_data = tblsmain_data;
+    options.tblsslim_index = tblsslim_index;
+    options.tblsslim_data = tblsslim_data;
     options.expire_tiles_zoom = expire_tiles_zoom;
     options.expire_tiles_zoom_min = expire_tiles_zoom_min;
     options.expire_tiles_filename = expire_tiles_filename;
@@ -479,6 +582,11 @@ int main(int argc, char *argv[])
     options.hstore_columns = hstore_columns;
     options.n_hstore_columns = n_hstore_columns;
     options.keep_coastlines = keep_coastlines;
+    options.parallel_indexing = parallel_indexing;
+    options.alloc_chunkwise = alloc_chunkwise;
+    options.num_procs = num_procs;
+    options.droptemp = droptemp;
+    options.unlogged = unlogged;
 
     if (strcmp("pgsql", output_backend) == 0) {
       osmdata.out = &out_pgsql;
@@ -490,6 +598,7 @@ int main(int argc, char *argv[])
       fprintf(stderr, "Output backend `%s' not recognised. Should be one of [pgsql, gazetteer, null].\n", output_backend);
       exit(EXIT_FAILURE);
     }
+    options.out = osmdata.out;
 
     if (strcmp("auto", input_reader) != 0) {
       if (strcmp("libxml2", input_reader) == 0) {
@@ -510,6 +619,9 @@ int main(int argc, char *argv[])
       }
     }
 
+    time_t overall_start, overall_end;
+    time(&overall_start);
+
     osmdata.out->start(&options);
 
     realloc_nodes(&osmdata);
@@ -551,10 +663,18 @@ int main(int argc, char *argv[])
     xmlMemoryDump();
     
     if (osmdata.count_node || osmdata.count_way || osmdata.count_rel) {
+        time_t now;
+        time(&now);
+        time_t end_nodes = osmdata.start_way > 0 ? osmdata.start_way : now;
+        time_t end_way = osmdata.start_rel > 0 ? osmdata.start_rel : now;
+        time_t end_rel =  now;
         fprintf(stderr, "\n");
-        fprintf(stderr, "Node stats: total(%d), max(%d)\n", osmdata.count_node, osmdata.max_node);
-        fprintf(stderr, "Way stats: total(%d), max(%d)\n", osmdata.count_way, osmdata.max_way);
-        fprintf(stderr, "Relation stats: total(%d), max(%d)\n", osmdata.count_rel, osmdata.max_rel);
+        fprintf(stderr, "Node stats: total(%" PRIdOSMID "), max(%" PRIdOSMID ") in %is\n", osmdata.count_node, osmdata.max_node,
+                osmdata.count_node > 0 ? (int)(end_nodes - osmdata.start_node) : 0);
+        fprintf(stderr, "Way stats: total(%" PRIdOSMID "), max(%" PRIdOSMID ") in %is\n", osmdata.count_way, osmdata.max_way,
+                osmdata.count_way > 0 ? (int)(end_way - osmdata.start_way) : 0);
+        fprintf(stderr, "Relation stats: total(%" PRIdOSMID "), max(%" PRIdOSMID ") in %is\n", osmdata.count_rel, osmdata.max_rel,
+                osmdata.count_rel > 0 ? (int)(end_rel - osmdata.start_rel) : 0);
     }
     osmdata.out->stop();
     
@@ -567,6 +687,8 @@ int main(int argc, char *argv[])
     project_exit();
     text_exit();
     fprintf(stderr, "\n");
+    time(&overall_end);
+    fprintf(stderr, "Osm2pgsql took %ds overall\n", (int)(overall_end - overall_start));
 
     return 0;
 }
