@@ -14,6 +14,7 @@
 #include <string.h>
 #include <assert.h>
 #include <errno.h>
+#include <time.h>
 
 #ifdef HAVE_PTHREAD
 #include <pthread.h>
@@ -117,9 +118,9 @@ static struct {
 };
 static const unsigned int nLayers = (sizeof(layers)/sizeof(*layers));
 
-static int pgsql_delete_way_from_output(int osm_id);
-static int pgsql_delete_relation_from_output(int osm_id);
-static int pgsql_process_relation(int id, struct member *members, int member_count, struct keyval *tags, int exists);
+static int pgsql_delete_way_from_output(osmid_t osm_id);
+static int pgsql_delete_relation_from_output(osmid_t osm_id);
+static int pgsql_process_relation(osmid_t id, struct member *members, int member_count, struct keyval *tags, int exists);
 
 void read_style_file( const char *filename )
 {
@@ -180,7 +181,7 @@ void read_style_file( const char *filename )
         fprintf( stderr, "Unknown flag '%s' line %d, ignored\n", str, lineno );
     }
     if (temp.flags==FLAG_PHSTORE) {
-      if (0==(Options->enable_hstore)) {
+      if (HSTORE_NONE==(Options->enable_hstore)) {
 	fprintf( stderr, "Error reading style file line %d (fields=%d)\n", lineno, fields );
 	fprintf( stderr, "flag 'phstore' is invalid in non-hstore mode\n");
 	exit_nicely();
@@ -421,21 +422,62 @@ static void pgsql_out_cleanup(void)
 
 /* Escape data appropriate to the type */
 static void escape_type(char *sql, int len, const char *value, const char *type) {
-    int items, from, to;
+  int items;
+  static int tmplen=0;
+  static char *tmpstr;
 
-    if ( !strcmp(type, "int4") ) {
-        /* For integers we take the first number, or the average if it's a-b */
-        items = sscanf(value, "%d-%d", &from, &to);
-        if ( items == 1 ) {
-            sprintf(sql, "%d", from);
-        } else if ( items == 2 ) {
-            sprintf(sql, "%d", (from + to) / 2);
-        } else {
-            sprintf(sql, "\\N");
-        }
+  if (len > tmplen) {
+    tmpstr=realloc(tmpstr,len);
+    tmplen=len;
+  }
+  strcpy(tmpstr,value);
+
+  if ( !strcmp(type, "int4") ) {
+    int from, to; 
+    /* For integers we take the first number, or the average if it's a-b */
+    items = sscanf(value, "%d-%d", &from, &to);
+    if ( items == 1 ) {
+      sprintf(sql, "%d", from);
+    } else if ( items == 2 ) {
+      sprintf(sql, "%d", (from + to) / 2);
     } else {
-        escape(sql, len, value);
+      sprintf(sql, "\\N");
     }
+  } else {
+    /*
+    try to "repair" real values as follows:
+      * assume "," to be a decimal mark which need to be replaced by "."
+      * like int4 take the first number, or the average if it's a-b
+      * assume SI unit (meters)
+      * convert feet to meters (1 foot = 0.3048 meters)
+      * reject anything else    
+    */
+    if ( !strcmp(type, "real") ) {
+      int i,slen;
+      float from,to;
+
+      slen=strlen(value);
+      for (i=0;i<slen;i++) if (tmpstr[i]==',') tmpstr[i]='.';
+
+      items = sscanf(tmpstr, "%f-%f", &from, &to);
+      if ( items == 1 ) {
+	if ((tmpstr[slen-2]=='f') && (tmpstr[slen-1]=='t')) {
+	  from*=0.3048;
+	}
+	sprintf(sql, "%f", from);
+      } else if ( items == 2 ) {
+	if ((tmpstr[slen-2]=='f') && (tmpstr[slen-1]=='t')) {
+	  from*=0.3048;
+	  to*=0.3048;
+	}
+	sprintf(sql, "%f", (from + to) / 2);
+      } else {
+	sprintf(sql, "\\N");
+      }
+    } else {
+      escape(sql, len, value);
+    }
+  }
 }
 
 static void write_hstore(enum table_id table, struct keyval *tags)
@@ -455,34 +497,42 @@ static void write_hstore(enum table_id table, struct keyval *tags)
     // while this tags has a follow-up..
     while (xtags->next->key != NULL)
     {
-        /*
-          hstore ASCII representation looks like
-          "<key>"=>"<value>"
-          
-          we need at least strlen(key)+strlen(value)+6+'\0' bytes
-          in theory any single character could also be escaped
-          thus we need an additional factor of 2.
-          The maximum lenght of a single hstore element is thus
-          calcuated as follows:
-        */
-        size_t hlen=2 * (strlen(xtags->next->key) + strlen(xtags->next->value)) + 7;
+
+      /* hard exclude z_order tag and keys which have their own column */
+      if ((xtags->next->has_column) || (strcmp("z_order",xtags->next->key)==0)) {
+	// update the tag-pointer to point to the next tag
+	xtags = xtags->next;
+	continue;
+      }
+
+      /*
+	hstore ASCII representation looks like
+	"<key>"=>"<value>"
         
-        // if the sql buffer is too small
-        if (hlen > sqllen) {
-            sqllen = hlen;
-            sql = realloc(sql, sqllen);
-        }
+	we need at least strlen(key)+strlen(value)+6+'\0' bytes
+	in theory any single character could also be escaped
+	thus we need an additional factor of 2.
+	The maximum lenght of a single hstore element is thus
+	calcuated as follows:
+      */
+      size_t hlen=2 * (strlen(xtags->next->key) + strlen(xtags->next->value)) + 7;
+      
+      // if the sql buffer is too small
+      if (hlen > sqllen) {
+	sqllen = hlen;
+	sql = realloc(sql, sqllen);
+      }
         
-        // pack the tag with its value into the hstore
-        keyval2hstore(sql, xtags->next);
-        copy_to_table(table, sql);
+      // pack the tag with its value into the hstore
+      keyval2hstore(sql, xtags->next);
+      copy_to_table(table, sql);
+
+      // update the tag-pointer to point to the next tag
+      xtags = xtags->next;
         
-        // update the tag-pointer to point to the next tag
-        xtags = xtags->next;
-        
-        // if the tag has a follow up, add a comma to the end
-        if (xtags->next->key != NULL)
-            copy_to_table(table, ",");
+      // if the tag has a follow up, add a comma to the end
+      if (xtags->next->key != NULL)
+	copy_to_table(table, ",");
     }
     
     // finish the hstore column by placing a TAB into the data stream
@@ -583,13 +633,14 @@ psql - 01 01000020 E6100000 30CCA462B6C3D4BF92998C9B38E04940
 Workaround - output SRID=4326;<WKB>
 */
 
-static int pgsql_out_node(int id, struct keyval *tags, double node_lat, double node_lon)
+static int pgsql_out_node(osmid_t id, struct keyval *tags, double node_lat, double node_lon)
 {
 
     static char *sql;
     static size_t sqllen=0;
     char *v;
     int i;
+    struct keyval *tag;
 
     if (sqllen==0) {
       sqllen=2048;
@@ -597,7 +648,7 @@ static int pgsql_out_node(int id, struct keyval *tags, double node_lat, double n
     }
 
     expire_tiles_from_bbox(node_lon, node_lat, node_lon, node_lat);
-    sprintf(sql, "%d\t", id);
+    sprintf(sql, "%" PRIdOSMID "\t", id);
     copy_to_table(t_point, sql);
 
     for (i=0; i < exportListCount[OSMTYPE_NODE]; i++) {
@@ -605,10 +656,12 @@ static int pgsql_out_node(int id, struct keyval *tags, double node_lat, double n
             continue;
 	if( (exportList[OSMTYPE_NODE][i].flags & FLAG_PHSTORE) == FLAG_PHSTORE)
             continue;
-        if ((v = getItem(tags, exportList[OSMTYPE_NODE][i].name)))
+        if ((tag = getTag(tags, exportList[OSMTYPE_NODE][i].name)))
         {
-            escape_type(sql, sqllen, v, exportList[OSMTYPE_NODE][i].type);
+            escape_type(sql, sqllen, tag->value, exportList[OSMTYPE_NODE][i].type);
             exportList[OSMTYPE_NODE][i].count++;
+	    if (HSTORE_NORM==Options->enable_hstore)
+	      tag->has_column=1;
         }
         else
             sprintf(sql, "\\N");
@@ -633,20 +686,21 @@ static int pgsql_out_node(int id, struct keyval *tags, double node_lat, double n
 
 
 
-static void write_wkts(int id, struct keyval *tags, const char *wkt, enum table_id table)
+static void write_wkts(osmid_t id, struct keyval *tags, const char *wkt, enum table_id table)
 {
   
     static char *sql;
     static size_t sqllen=0;
     char *v;
     int j;
+    struct keyval *tag;
 
     if (sqllen==0) {
       sqllen=2048;
       sql=malloc(sqllen);
     }
     
-    sprintf(sql, "%d\t", id);
+    sprintf(sql, "%" PRIdOSMID "\t", id);
     copy_to_table(table, sql);
 
     for (j=0; j < exportListCount[OSMTYPE_WAY]; j++) {
@@ -654,10 +708,12 @@ static void write_wkts(int id, struct keyval *tags, const char *wkt, enum table_
                 continue;
 	    if( (exportList[OSMTYPE_WAY][j].flags & FLAG_PHSTORE) == FLAG_PHSTORE)
                 continue;
-            if ((v = getItem(tags, exportList[OSMTYPE_WAY][j].name)))
+            if ((tag = getTag(tags, exportList[OSMTYPE_WAY][j].name)))
             {
                 exportList[OSMTYPE_WAY][j].count++;
-                escape_type(sql, sqllen, v, exportList[OSMTYPE_WAY][j].type);
+                escape_type(sql, sqllen, tag->value, exportList[OSMTYPE_WAY][j].type);
+		if (HSTORE_NORM==Options->enable_hstore)
+		  tag->has_column=1;
             }
             else
                 sprintf(sql, "\\N");
@@ -786,7 +842,7 @@ E4C1421D5BF24D06053E7DF4940
 212696  Oswald Road     \N      \N      \N      \N      \N      \N      minor   \N      \N      \N      \N      \N      \N      \N    0102000020E610000004000000467D923B6C22D5BFA359D93EE4DF4940B3976DA7AD11D5BF84BBB376DBDF4940997FF44D9A06D5BF4223D8B8FEDF49404D158C4AEA04D
 5BF5BB39597FCDF4940
 */
-static int pgsql_out_way(int id, struct keyval *tags, struct osmNode *nodes, int count, int exists)
+static int pgsql_out_way(osmid_t id, struct keyval *tags, struct osmNode *nodes, int count, int exists)
 {
     int polygon = 0, roads = 0;
     int i, wkt_size;
@@ -838,7 +894,7 @@ static int pgsql_out_way(int id, struct keyval *tags, struct osmNode *nodes, int
     return 0;
 }
 
-static int pgsql_out_relation(int id, struct keyval *rel_tags, struct osmNode **xnodes, struct keyval *xtags, int *xcount, int *xid, const char **xrole)
+static int pgsql_out_relation(osmid_t id, struct keyval *rel_tags, struct osmNode **xnodes, struct keyval *xtags, int *xcount, osmid_t *xid, const char **xrole)
 {
     int i, wkt_size;
     int polygon = 0, roads = 0;
@@ -1123,6 +1179,26 @@ static int pgsql_out_relation(int id, struct keyval *rel_tags, struct osmNode **
     return 0;
 }
 
+static int pgsql_out_connect(const struct output_options *options, int startTransaction) {
+    int i;
+    for (i=0; i<NUM_TABLES; i++) {
+        PGconn *sql_conn;
+        sql_conn = PQconnectdb(options->conninfo);
+        
+        /* Check to see that the backend connection was successfully made */
+        if (PQstatus(sql_conn) != CONNECTION_OK) {
+            fprintf(stderr, "Connection to database failed: %s\n", PQerrorMessage(sql_conn));
+            exit_nicely();
+        }
+        tables[i].sql_conn = sql_conn;
+        pgsql_exec(sql_conn, PGRES_COMMAND_OK, "SET synchronous_commit TO off;");
+        pgsql_exec(sql_conn, PGRES_COMMAND_OK, "PREPARE get_wkt (" POSTGRES_OSMID_TYPE ") AS SELECT ST_AsText(way) FROM %s WHERE osm_id = $1;\n", tables[i].name);
+        if (startTransaction)
+            pgsql_exec(sql_conn, PGRES_COMMAND_OK, "BEGIN");
+    }
+    return 0;
+}
+
 static int pgsql_out_start(const struct output_options *options)
 {
     char *sql, tmp[256];
@@ -1156,6 +1232,7 @@ static int pgsql_out_start(const struct output_options *options)
             exit_nicely();
         }
         tables[i].sql_conn = sql_conn;
+        pgsql_exec(sql_conn, PGRES_COMMAND_OK, "SET synchronous_commit TO off;");
 
         if (!options->append) {
             pgsql_exec(sql_conn, PGRES_COMMAND_OK, "DROP TABLE IF EXISTS %s", tables[i].name);
@@ -1187,12 +1264,12 @@ static int pgsql_out_start(const struct output_options *options)
         int numTags = exportListCount[type];
         struct taginfo *exportTags = exportList[type];
         if (!options->append) {
-            sprintf(sql, "CREATE TABLE %s ( osm_id int4", tables[i].name );
+            sprintf(sql, "CREATE TABLE %s ( osm_id " POSTGRES_OSMID_TYPE, tables[i].name );
             for (j=0; j < numTags; j++) {
                 if( exportTags[j].flags & FLAG_DELETE )
                     continue;
-		if( (exportTags[j].flags & FLAG_PHSTORE ) == FLAG_PHSTORE)
-		    continue;
+                if( (exportTags[j].flags & FLAG_PHSTORE ) == FLAG_PHSTORE)
+                    continue;
                 sprintf(tmp, ",\"%s\" %s", exportTags[j].name, exportTags[j].type);
                 if (strlen(sql) + strlen(tmp) + 1 > sql_len) {
                     sql_len *= 2;
@@ -1209,18 +1286,26 @@ static int pgsql_out_start(const struct output_options *options)
                 strcat(sql, "\" hstore ");
             }
             if (Options->enable_hstore) {
-                strcat(sql, ",tags hstore );\n");
-	    } else {
-	      strcat(sql, " );\n");
-	    }
+                strcat(sql, ",tags hstore");
+            } 
+            strcat(sql, ")");
+            if (Options->tblsmain_data) {
+                sprintf(sql + strlen(sql), " TABLESPACE %s", Options->tblsmain_data);
+            }
+            strcat(sql, "\n");
 
             pgsql_exec(sql_conn, PGRES_COMMAND_OK, "%s", sql);
             pgsql_exec(sql_conn, PGRES_TUPLES_OK, "SELECT AddGeometryColumn('%s', 'way', %d, '%s', 2 );\n",
                         tables[i].name, SRID, tables[i].type );
             pgsql_exec(sql_conn, PGRES_COMMAND_OK, "ALTER TABLE %s ALTER COLUMN way SET NOT NULL;\n", tables[i].name);
             /* slim mode needs this to be able to apply diffs */
-            if( Options->slim )
-	      pgsql_exec(sql_conn, PGRES_COMMAND_OK, "CREATE INDEX %s_pkey ON %s USING BTREE (osm_id) TABLESPACE %s;\n", tables[i].name, tables[i].name, Options->tblsindex);
+            if (Options->slim && !Options->droptemp) {
+                sprintf(sql, "CREATE INDEX %s_pkey ON %s USING BTREE (osm_id)",  tables[i].name, tables[i].name);
+                if (Options->tblsmain_index) {
+                    sprintf(sql + strlen(sql), " TABLESPACE %s\n", Options->tblsmain_index);
+                }
+	            pgsql_exec(sql_conn, PGRES_COMMAND_OK, "%s", sql);
+            }
         } else {
             /* Add any new columns referenced in the default.style */
             PGresult *res;
@@ -1251,10 +1336,10 @@ static int pgsql_out_start(const struct output_options *options)
             PQclear(res);
 
             /* change the type of the geometry column if needed - this can only change to a more permisive type */
-            pgsql_exec(sql_conn, PGRES_COMMAND_OK, "UPDATE geometry_columns SET type = '%s' where type != '%s' and f_table_name = '%s' and f_geometry_column = 'way'",
-                        tables[i].type, tables[i].type, tables[i].name);
+            //            pgsql_exec(sql_conn, PGRES_COMMAND_OK, "UPDATE geometry_columns SET type = '%s' where type != '%s' and f_table_name = '%s' and f_geometry_column = 'way'",
+            //           tables[i].type, tables[i].type, tables[i].name);
         }
-        pgsql_exec(sql_conn, PGRES_COMMAND_OK, "PREPARE get_way (int4) AS SELECT AsText(way) FROM %s WHERE osm_id = $1;\n", tables[i].name);
+        pgsql_exec(sql_conn, PGRES_COMMAND_OK, "PREPARE get_wkt (" POSTGRES_OSMID_TYPE ") AS SELECT ST_AsText(way) FROM %s WHERE osm_id = $1;\n", tables[i].name);
         
         /* Generate column list for COPY */
         strcpy(sql, "osm_id");
@@ -1320,6 +1405,28 @@ static void pgsql_pause_copy(struct s_table *table)
     table->copyMode = 0;
 }
 
+static void pgsql_out_close(int stopTransaction) {
+    int i;
+    for (i=0; i<NUM_TABLES; i++) {
+        pgsql_pause_copy(&tables[i]);
+        // Commit transaction
+        if (stopTransaction)
+            pgsql_exec(tables[i].sql_conn, PGRES_COMMAND_OK, "COMMIT");
+        PQfinish(tables[i].sql_conn);
+        tables[i].sql_conn = NULL;
+    }
+}
+
+static void pgsql_out_commit(void) {
+    int i;
+    for (i=0; i<NUM_TABLES; i++) {
+        pgsql_pause_copy(&tables[i]);
+        // Commit transaction
+        fprintf(stderr, "Committing transaction for %s\n", tables[i].name);
+        pgsql_exec(tables[i].sql_conn, PGRES_COMMAND_OK, "COMMIT");
+    }
+}
+
 static void *pgsql_out_stop_one(void *arg)
 {
     struct s_table *table = arg;
@@ -1333,29 +1440,55 @@ static void *pgsql_out_stop_one(void *arg)
 
     pgsql_pause_copy(table);
     // Commit transaction
-    fprintf(stderr, "Committing transaction for %s\n", table->name);
-    pgsql_exec(sql_conn, PGRES_COMMAND_OK, "COMMIT");
-    if( !Options->append )
+    //fprintf(stderr, "Committing transaction for %s\n", table->name);
+    //pgsql_exec(sql_conn, PGRES_COMMAND_OK, "COMMIT");
+    if (!Options->append)
     {
-      fprintf(stderr, "Sorting data and creating indexes for %s\n", table->name);
-      pgsql_exec(sql_conn, PGRES_COMMAND_OK, "ANALYZE %s;\n", table->name);
-      pgsql_exec(sql_conn, PGRES_COMMAND_OK, "CREATE TABLE %s_tmp AS SELECT * FROM %s ORDER BY way;\n", table->name, table->name);
-      pgsql_exec(sql_conn, PGRES_COMMAND_OK, "DROP TABLE %s;\n", table->name);
-      pgsql_exec(sql_conn, PGRES_COMMAND_OK, "ALTER TABLE %s_tmp RENAME TO %s;\n", table->name, table->name);
-      pgsql_exec(sql_conn, PGRES_COMMAND_OK, "CREATE INDEX %s_index ON %s USING GIST (way GIST_GEOMETRY_OPS) TABLESPACE %s;\n", table->name, table->name, Options->tblsindex);
-      /* slim mode needs this to be able to apply diffs */
-      if( Options->slim )
-	pgsql_exec(sql_conn, PGRES_COMMAND_OK, "CREATE INDEX %s_pkey ON %s USING BTREE (osm_id) TABLESPACE %s;\n", table->name, table->name, Options->tblsindex);
-      pgsql_exec(sql_conn, PGRES_COMMAND_OK, "GRANT SELECT ON %s TO PUBLIC;\n", table->name);
-      pgsql_exec(sql_conn, PGRES_COMMAND_OK, "ANALYZE %s;\n", table->name);
+        time_t start, end;
+        time(&start);
+        fprintf(stderr, "Sorting data and creating indexes for %s\n", table->name);
+        pgsql_exec(sql_conn, PGRES_COMMAND_OK, "ANALYZE %s;\n", table->name);
+        fprintf(stderr, "Analyzing %s finished\n", table->name);
+        if (Options->tblsmain_data) {
+            pgsql_exec(sql_conn, PGRES_COMMAND_OK, "CREATE TABLE %s_tmp "
+                        "TABLESPACE %s AS SELECT * FROM %s ORDER BY way;\n",
+                        table->name, Options->tblsmain_data, table->name);
+        } else {
+            pgsql_exec(sql_conn, PGRES_COMMAND_OK, "CREATE TABLE %s_tmp AS SELECT * FROM %s ORDER BY way;\n", table->name, table->name);
+        }
+        pgsql_exec(sql_conn, PGRES_COMMAND_OK, "DROP TABLE %s;\n", table->name);
+        pgsql_exec(sql_conn, PGRES_COMMAND_OK, "ALTER TABLE %s_tmp RENAME TO %s;\n", table->name, table->name);
+        fprintf(stderr, "Copying %s to cluster by geometry finished\n", table->name);
+        if (Options->tblsmain_index) {
+            pgsql_exec(sql_conn, PGRES_COMMAND_OK, "CREATE INDEX %s_index ON %s USING GIST (way) TABLESPACE %s;\n", table->name, table->name, Options->tblsmain_index);
+        } else {
+            pgsql_exec(sql_conn, PGRES_COMMAND_OK, "CREATE INDEX %s_index ON %s USING GIST (way);\n", table->name, table->name);
+        }
+
+        /* slim mode needs this to be able to apply diffs */
+        if (Options->slim && !Options->droptemp)
+        {
+            if (Options->tblsmain_index) {
+                pgsql_exec(sql_conn, PGRES_COMMAND_OK, "CREATE INDEX %s_pkey ON %s USING BTREE (osm_id) TABLESPACE %s;\n", table->name, table->name, Options->tblsmain_index);
+            } else {
+                pgsql_exec(sql_conn, PGRES_COMMAND_OK, "CREATE INDEX %s_pkey ON %s USING BTREE (osm_id);\n", table->name, table->name);
+            }
+        }
+        fprintf(stderr, "Creating indexes on  %s finished\n", table->name);
+        pgsql_exec(sql_conn, PGRES_COMMAND_OK, "GRANT SELECT ON %s TO PUBLIC;\n", table->name);
+        pgsql_exec(sql_conn, PGRES_COMMAND_OK, "ANALYZE %s;\n", table->name);
+        time(&end);
+        fprintf(stderr, "All indexes on  %s created  in %ds\n", table->name, (int)(end - start));
     }
     PQfinish(sql_conn);
     table->sql_conn = NULL;
+
     fprintf(stderr, "Completed %s\n", table->name);
     free(table->name);
     free(table->columns);
     return NULL;
 }
+
 static void pgsql_out_stop()
 {
     int i;
@@ -1363,43 +1496,73 @@ static void pgsql_out_stop()
     pthread_t threads[NUM_TABLES];
 #endif
 
+    /* Commit the transactions, so that multiple processes can
+     * access the data simultanious to process the rest in parallel
+     * as well as see the newly created tables.
+     */
+    pgsql_out_commit();
+    Options->mid->commit();
+    /* To prevent deadlocks in parallel processing, the mid tables need
+     * to stay out of a transaction. In this stage output tables are only
+     * written to and not read, so they can be processed as several parallel
+     * independent transactions
+     */
+    for (i=0; i<NUM_TABLES; i++) {
+        PGconn *sql_conn = tables[i].sql_conn;
+        pgsql_exec(sql_conn, PGRES_COMMAND_OK, "BEGIN");
+    }
     /* Processing any remaing to be processed ways */
     Options->mid->iterate_ways( pgsql_out_way );
+    pgsql_out_commit();
+    Options->mid->commit();
+
+    /* Processing any remaing to be processed relations */
+    /* During this stage output tables also need to stay out of
+     * extended transactions, as the delete_way_from_output, called
+     * from process_relation, can deadlock if using multi-processing.
+     */    
     Options->mid->iterate_relations( pgsql_process_relation );
 
 #ifdef HAVE_PTHREAD
-    for (i=0; i<NUM_TABLES; i++) {
-        int ret = pthread_create(&threads[i], NULL, pgsql_out_stop_one, &tables[i]);
-        if (ret) {
-            fprintf(stderr, "pthread_create() returned an error (%d)", ret);
-            exit_nicely();
-        }
-    }
+    if (Options->parallel_indexing) {
+      for (i=0; i<NUM_TABLES; i++) {
+          int ret = pthread_create(&threads[i], NULL, pgsql_out_stop_one, &tables[i]);
+          if (ret) {
+              fprintf(stderr, "pthread_create() returned an error (%d)", ret);
+              exit_nicely();
+          }
+      }
+  
+      /* No longer need to access middle layer -- release memory */
+      Options->mid->stop();
+  
+      for (i=0; i<NUM_TABLES; i++) {
+          int ret = pthread_join(threads[i], NULL);
+          if (ret) {
+              fprintf(stderr, "pthread_join() returned an error (%d)", ret);
+              exit_nicely();
+          }
+      }
+    } else {
+#endif
 
-    /* No longer need to access middle layer -- release memory */
-    Options->mid->stop();
-
-    for (i=0; i<NUM_TABLES; i++) {
-        int ret = pthread_join(threads[i], NULL);
-        if (ret) {
-            fprintf(stderr, "pthread_join() returned an error (%d)", ret);
-            exit_nicely();
-        }
-    }
-#else
     /* No longer need to access middle layer -- release memory */
     Options->mid->stop();
     for (i=0; i<NUM_TABLES; i++)
         pgsql_out_stop_one(&tables[i]);
+
+#ifdef HAVE_PTHREAD
+    }
 #endif
 
+
     pgsql_out_cleanup();
     free_style();
 
     expire_tiles_stop();
 }
 
-static int pgsql_add_node(int id, double lat, double lon, struct keyval *tags)
+static int pgsql_add_node(osmid_t id, double lat, double lon, struct keyval *tags)
 {
   int polygon;
   int filter = pgsql_filter_tags(OSMTYPE_NODE, tags, &polygon);
@@ -1410,7 +1573,7 @@ static int pgsql_add_node(int id, double lat, double lon, struct keyval *tags)
   return 0;
 }
 
-static int pgsql_add_way(int id, int *nds, int nd_count, struct keyval *tags)
+static int pgsql_add_way(osmid_t id, osmid_t *nds, int nd_count, struct keyval *tags)
 {
   int polygon = 0;
 
@@ -1433,11 +1596,11 @@ static int pgsql_add_way(int id, int *nds, int nd_count, struct keyval *tags)
 }
 
 /* This is the workhorse of pgsql_add_relation, split out because it is used as the callback for iterate relations */
-static int pgsql_process_relation(int id, struct member *members, int member_count, struct keyval *tags, int exists)
+static int pgsql_process_relation(osmid_t id, struct member *members, int member_count, struct keyval *tags, int exists)
 {
-  // (int id, struct keyval *rel_tags, struct osmNode **xnodes, struct keyval **xtags, int *xcount)
+  // (osmid_t id, struct keyval *rel_tags, struct osmNode **xnodes, struct keyval **xtags, int *xcount)
   int i, count;
-  int *xid = malloc( (member_count+1) * sizeof(int) );
+  osmid_t *xid = malloc( (member_count+1) * sizeof(osmid_t) );
   const char **xrole = malloc( (member_count+1) * sizeof(const char *) );
   int *xcount = malloc( (member_count+1) * sizeof(int) );
   struct keyval *xtags  = malloc( (member_count+1) * sizeof(struct keyval) );
@@ -1465,7 +1628,6 @@ static int pgsql_process_relation(int id, struct member *members, int member_cou
   xcount[count] = 0;
   xid[count] = 0;
   xrole[count] = NULL;
-
   // At some point we might want to consider storing the retreived data in the members, rather than as seperate arrays
   pgsql_out_relation(id, tags, xnodes, xtags, xcount, xid, xrole);
 
@@ -1483,7 +1645,7 @@ static int pgsql_process_relation(int id, struct member *members, int member_cou
   return 0;
 }
 
-static int pgsql_add_relation(int id, struct member *members, int member_count, struct keyval *tags)
+static int pgsql_add_relation(osmid_t id, struct member *members, int member_count, struct keyval *tags)
 {
   const char *type = getItem(tags, "type");
 
@@ -1494,7 +1656,7 @@ static int pgsql_add_relation(int id, struct member *members, int member_count,
   /* In slim mode we remember these*/
   if(Options->mid->relations_set)
     Options->mid->relations_set(id, members, member_count, tags);
-  // (int id, struct keyval *rel_tags, struct osmNode **xnodes, struct keyval **xtags, int *xcount)
+  // (osmid_t id, struct keyval *rel_tags, struct osmNode **xnodes, struct keyval **xtags, int *xcount)
 
   return pgsql_process_relation(id, members, member_count, tags, 0);
 }
@@ -1503,7 +1665,7 @@ static int pgsql_add_relation(int id, struct member *members, int member_count,
 /* Delete is easy, just remove all traces of this object. We don't need to
  * worry about finding objects that depend on it, since the same diff must
  * contain the change for that also. */
-static int pgsql_delete_node(int osm_id)
+static int pgsql_delete_node(osmid_t osm_id)
 {
     if( !Options->slim )
     {
@@ -1511,31 +1673,34 @@ static int pgsql_delete_node(int osm_id)
         exit_nicely();
     }
     pgsql_pause_copy(&tables[t_point]);
-    expire_tiles_from_db(tables[t_point].sql_conn, osm_id);
-    pgsql_exec(tables[t_point].sql_conn, PGRES_COMMAND_OK, "DELETE FROM %s WHERE osm_id = %d", tables[t_point].name, osm_id );
+    if ( expire_tiles_from_db(tables[t_point].sql_conn, osm_id) != 0)
+        pgsql_exec(tables[t_point].sql_conn, PGRES_COMMAND_OK, "DELETE FROM %s WHERE osm_id = %" PRIdOSMID, tables[t_point].name, osm_id );
+    
     Options->mid->nodes_delete(osm_id);
     return 0;
 }
 
 /* Seperated out because we use it elsewhere */
-static int pgsql_delete_way_from_output(int osm_id)
+static int pgsql_delete_way_from_output(osmid_t osm_id)
 {
     /* Optimisation: we only need this is slim mode */
     if( !Options->slim )
         return 0;
+    /* in droptemp mode we don't have indices and this takes ages. */
+    if (Options->droptemp)
+        return 0;
     pgsql_pause_copy(&tables[t_roads]);
     pgsql_pause_copy(&tables[t_line]);
     pgsql_pause_copy(&tables[t_poly]);
-    expire_tiles_from_db(tables[t_roads].sql_conn, osm_id);
-    expire_tiles_from_db(tables[t_line].sql_conn, osm_id);
-    expire_tiles_from_db(tables[t_poly].sql_conn, osm_id);
-    pgsql_exec(tables[t_roads].sql_conn, PGRES_COMMAND_OK, "DELETE FROM %s WHERE osm_id = %d", tables[t_roads].name, osm_id );
-    pgsql_exec(tables[t_line].sql_conn, PGRES_COMMAND_OK, "DELETE FROM %s WHERE osm_id = %d", tables[t_line].name, osm_id );
-    pgsql_exec(tables[t_poly].sql_conn, PGRES_COMMAND_OK, "DELETE FROM %s WHERE osm_id = %d", tables[t_poly].name, osm_id );
+    pgsql_exec(tables[t_roads].sql_conn, PGRES_COMMAND_OK, "DELETE FROM %s WHERE osm_id = %" PRIdOSMID, tables[t_roads].name, osm_id );
+    if ( expire_tiles_from_db(tables[t_line].sql_conn, osm_id) != 0)
+        pgsql_exec(tables[t_line].sql_conn, PGRES_COMMAND_OK, "DELETE FROM %s WHERE osm_id = %" PRIdOSMID, tables[t_line].name, osm_id );
+    if ( expire_tiles_from_db(tables[t_poly].sql_conn, osm_id) != 0)
+        pgsql_exec(tables[t_poly].sql_conn, PGRES_COMMAND_OK, "DELETE FROM %s WHERE osm_id = %" PRIdOSMID, tables[t_poly].name, osm_id );
     return 0;
 }
 
-static int pgsql_delete_way(int osm_id)
+static int pgsql_delete_way(osmid_t osm_id)
 {
     if( !Options->slim )
     {
@@ -1548,21 +1713,20 @@ static int pgsql_delete_way(int osm_id)
 }
 
 /* Relations are identified by using negative IDs */
-static int pgsql_delete_relation_from_output(int osm_id)
+static int pgsql_delete_relation_from_output(osmid_t osm_id)
 {
     pgsql_pause_copy(&tables[t_roads]);
     pgsql_pause_copy(&tables[t_line]);
     pgsql_pause_copy(&tables[t_poly]);
-    expire_tiles_from_db(tables[t_roads].sql_conn, -osm_id);
-    expire_tiles_from_db(tables[t_line].sql_conn, -osm_id);
-    expire_tiles_from_db(tables[t_poly].sql_conn, -osm_id);
-    pgsql_exec(tables[t_roads].sql_conn, PGRES_COMMAND_OK, "DELETE FROM %s WHERE osm_id = %d", tables[t_roads].name, -osm_id );
-    pgsql_exec(tables[t_line].sql_conn, PGRES_COMMAND_OK, "DELETE FROM %s WHERE osm_id = %d", tables[t_line].name, -osm_id );
-    pgsql_exec(tables[t_poly].sql_conn, PGRES_COMMAND_OK, "DELETE FROM %s WHERE osm_id = %d", tables[t_poly].name, -osm_id );
+    pgsql_exec(tables[t_roads].sql_conn, PGRES_COMMAND_OK, "DELETE FROM %s WHERE osm_id = %" PRIdOSMID, tables[t_roads].name, -osm_id );
+    if ( expire_tiles_from_db(tables[t_line].sql_conn, -osm_id) != 0)
+        pgsql_exec(tables[t_line].sql_conn, PGRES_COMMAND_OK, "DELETE FROM %s WHERE osm_id = %" PRIdOSMID, tables[t_line].name, -osm_id );
+    if ( expire_tiles_from_db(tables[t_poly].sql_conn, -osm_id) != 0)
+        pgsql_exec(tables[t_poly].sql_conn, PGRES_COMMAND_OK, "DELETE FROM %s WHERE osm_id = %" PRIdOSMID, tables[t_poly].name, -osm_id );
     return 0;
 }
 
-static int pgsql_delete_relation(int osm_id)
+static int pgsql_delete_relation(osmid_t osm_id)
 {
     if( !Options->slim )
     {
@@ -1577,7 +1741,7 @@ static int pgsql_delete_relation(int osm_id)
 /* Modify is slightly trickier. The basic idea is we simply delete the
  * object and create it with the new parameters. Then we need to mark the
  * objects that depend on this one */
-static int pgsql_modify_node(int osm_id, double lat, double lon, struct keyval *tags)
+static int pgsql_modify_node(osmid_t osm_id, double lat, double lon, struct keyval *tags)
 {
     if( !Options->slim )
     {
@@ -1590,7 +1754,7 @@ static int pgsql_modify_node(int osm_id, double lat, double lon, struct keyval *
     return 0;
 }
 
-static int pgsql_modify_way(int osm_id, int *nodes, int node_count, struct keyval *tags)
+static int pgsql_modify_way(osmid_t osm_id, osmid_t *nodes, int node_count, struct keyval *tags)
 {
     if( !Options->slim )
     {
@@ -1603,7 +1767,7 @@ static int pgsql_modify_way(int osm_id, int *nodes, int node_count, struct keyva
     return 0;
 }
 
-static int pgsql_modify_relation(int osm_id, struct member *members, int member_count, struct keyval *tags)
+static int pgsql_modify_relation(osmid_t osm_id, struct member *members, int member_count, struct keyval *tags)
 {
     if( !Options->slim )
     {
@@ -1618,8 +1782,10 @@ static int pgsql_modify_relation(int osm_id, struct member *members, int member_
 
 struct output_t out_pgsql = {
         .start           = pgsql_out_start,
+        .connect         = pgsql_out_connect,
         .stop            = pgsql_out_stop,
         .cleanup         = pgsql_out_cleanup,
+        .close           = pgsql_out_close,
         .node_add        = pgsql_add_node,
         .way_add         = pgsql_add_way,
         .relation_add    = pgsql_add_relation,
